package be.iminds.ilabt.jfed.experimenter_gui.canvas.rspec;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.CanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.ExperimentCanvas;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.NewLinkTarget;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecLink;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import javafx.beans.InvalidationListener;
import javafx.beans.value.ChangeListener;
import javafx.collections.ListChangeListener;
import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.scene.Node;
import javafx.scene.control.Button;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Visualisation class for Rspec ifaceLinks.
 * <p>
 * These can have a connection to more than two nodes, and are thus a collection of
 * two OR MORE ifaceLinks to a central "link"-component
 * User: twalcari
 * Date: 11/13/13
 * Time: 9:22 AM
 */
public class RspecCanvasLink implements ListChangeListener<FXRspecInterface> {

    private static final Logger LOG = LoggerFactory.getLogger(RspecCanvasLink.class);

    private static final String LINK_CENTER_STYLECLASS = "link-center";
    private static final String IMPAIRED_LINK_CENTER_STYLECLASS = "impaired-link-center";
    private static final String SHARED_LINK_CENTER_STYLECLASS = "shared-link-center";
    @Nonnull private final FXRspecLink link;
    @Nonnull private final ModelRspecToCanvasBinding binding;
    @Nonnull private final ExperimentCanvas experimentCanvas;
    @Nonnull private final LinkCenter linkCenter;
    private final Map<FXRspecInterface, InterfaceLink> ifaceLinks = new HashMap<>();
    private final InvalidationListener layoutListener = observable -> updateLayout();

    private double centerX, centerY;

    public RspecCanvasLink(@Nonnull FXRspecLink link, @Nonnull ModelRspecToCanvasBinding binding) {
        this.link = link;
        this.binding = binding;
        this.experimentCanvas = binding.getExperimentCanvas();

        //create centernode
        this.linkCenter = new LinkCenter();
        experimentCanvas.addLinkCenter(linkCenter);


        //add ifaceLinks between center and FXRspecNode

        link.getInterfaces().forEach(this::addInterface);
        link.getInterfaces().addListener(this);

        //make the link's type change the look off all "interface lines"
        link.getLinkTypes().addListener((Change<? extends String> c) -> {
            for (CanvasLink cl : ifaceLinks.values()) {
                if (link.getLinkTypes().isEmpty()) {
                    cl.setLinkType("lan");
                } else {
                    cl.setLinkType(link.getLinkTypes().get(0));
                }
            }
        });

        link.getComponentManagerUrns().addListener((Change<? extends GeniUrn> c) -> updateStitched(c.getList()));
        updateStitched(link.getComponentManagerUrns());

        //register to canvas for autolayouting support
        experimentCanvas.registerRspecCanvasLink(this);
    }

    private void updateStitched(List<? extends GeniUrn> cms) {
        Set<GeniUrn> urns = new HashSet<>(cms);
        for (CanvasLink cl : ifaceLinks.values()) {
            cl.setStitched(urns.size() > 1);
        }
        System.out.println("updateStitched urns="+urns);
    }

    private void addInterface(@Nonnull FXRspecInterface iface) {
        assert (!ifaceLinks.containsKey(iface));

        RspecCanvasNode rspecCanvasNode = binding.getRspecCanvasNode(iface.getNode());
        assert rspecCanvasNode != null;
        if (rspecCanvasNode == null) {
            //this should not happen (not sure why), and no good way to recover if it does
            throw new RuntimeException("Could not add interface because no canvas node found in binding");
        }

        rspecCanvasNode.layoutXProperty().addListener(layoutListener);
        rspecCanvasNode.layoutYProperty().addListener(layoutListener);
        rspecCanvasNode.boundsInLocalProperty().addListener(layoutListener);


        InterfaceLink canvasLink = new InterfaceLink(iface, rspecCanvasNode, linkCenter);
        if (!link.getLinkTypes().isEmpty()) {
            canvasLink.setLinkType(link.getLinkTypes().get(0));
        }

        ifaceLinks.put(iface, canvasLink);

        //add as first element to make sure that ifaceLinks are rendered first, and are thus overriden by nodes
        experimentCanvas.addLink(canvasLink);

        updateLayout();
    }

    private void removeInterface(@Nonnull FXRspecInterface iface) {
        InterfaceLink canvasLink = ifaceLinks.get(iface);
        RspecCanvasNode rspecCanvasNode = canvasLink.getNodeA();

        rspecCanvasNode.layoutXProperty().removeListener(layoutListener);
        rspecCanvasNode.layoutYProperty().removeListener(layoutListener);
        rspecCanvasNode.boundsInLocalProperty().removeListener(layoutListener);

        experimentCanvas.removeLink(canvasLink);
        ifaceLinks.remove(iface);
        updateLayout();
    }

    private void updateLayout() {
        if (ifaceLinks.isEmpty()) return;

        centerX = 0;
        centerY = 0;
        for (InterfaceLink ntlcLink : ifaceLinks.values()) {
            RspecCanvasNode node = ntlcLink.getNodeA();

            centerX += NewLinkTarget.getCenterX(node);
            centerY += NewLinkTarget.getCenterY(node);
        }
        centerX /= ifaceLinks.size();
        centerY /= ifaceLinks.size();

        if (ifaceLinks.size() <= 1) {
            //make sure that broken links are visualized nicely
            centerY += 100;
        }

        NewLinkTarget.setCenterX(centerX, linkCenter);
        NewLinkTarget.setCenterY(centerY, linkCenter);
    }

    /**
     * Should be called when the RspecCanvasLink is removed from the canvas
     */
    public void remove() {
        //unregister from canvas (was for autolayouting support)
        experimentCanvas.unregisterRspecCanvasLink(this);

        experimentCanvas.removeLinkCenter(linkCenter);

        for (InterfaceLink ntlcLink : ifaceLinks.values()) {
            experimentCanvas.removeLink(ntlcLink);
        }

        link.getInterfaces().removeListener(this);
    }

    @Override
    public void onChanged(Change<? extends FXRspecInterface> change) {
        LOG.trace("Detected interface changes on link {}", link.getUniqueId());

        while (change.next()) {
            LOG.trace("Received interface change on link {}: {}", link.getUniqueId(), change);

            if (change.wasAdded()) {
                change.getAddedSubList().forEach(iface -> LOG.trace("iface {} was added to rspecCanvasLink {}", iface.getClientId(), getRspecLink().getClientId()));
                change.getAddedSubList().forEach(this::addInterface);
            }

            if (change.wasRemoved()) {
                change.getRemoved().forEach(iface -> LOG.trace("iface {} was removed to rspecCanvasLink {}", iface.getClientId(), getRspecLink().getClientId()));
                change.getRemoved().forEach(this::removeInterface);
            }
        }
    }

    public double getCenterX() {
        return centerX;
    }

    public double getCenterY() {
        return centerY;
    }

    @Nonnull
    public FXRspecLink getRspecLink() {
        return link;
    }

    @Nonnull
    public Button getLinkCenter() {
        return linkCenter;
    }

    @Nonnull
    public Collection<InterfaceLink> getInterfaceLinks() {
        return ifaceLinks.values();
    }

    @Nullable
    public InterfaceLink getInterfaceLinkByRspecInterface(@Nonnull FXRspecInterface iface) {
        return ifaceLinks.get(iface);
    }


    public class LinkCenter extends Button implements NewLinkTarget {

        private boolean hasImpairedLook = false;
        private boolean hasAutoSharedLook = false;
        private boolean hasAddToSharedLook = false;
        @Nonnull private final Node createSharedGraphic;
        @Nonnull private final Node addTosharedGraphic;

        public LinkCenter() {
            super();
            createSharedGraphic = GlyphUtils.createSmallGlyph(FontAwesome.Glyph.USERS, Color.BLACK);
            addTosharedGraphic = GlyphUtils.createSmallGlyph(FontAwesome.Glyph.USER_PLUS, Color.BLACK);
            //other icons: PLUG  CLOUD  CLOUD_UPLOAD  cLOUD_DOWNLOAD

            //was very simple: textProperty().bind(link.clientIdProperty());
            updateLinkText();
            link.clientIdProperty().addListener((ob, ov, nw) -> updateLinkText());
            link.getLinkTypes().addListener((Change<? extends String> c) -> updateLinkText());
            link.getComponentManagerUrns().addListener((Change<? extends GeniUrn> c) -> updateLinkText());

            getStyleClass().add(LINK_CENTER_STYLECLASS);
            setOnAction(actionEvent -> onLinkCenterClicked());
            layoutBoundsProperty().addListener(observable -> updateLayout());
            focusedProperty().addListener((observableValue, oldValue, newValue) -> {
                for (CanvasLink cl : ifaceLinks.values()) {
                    cl.setSelected(newValue);

                }
            });


            link.impairedProperty().addListener(observable -> {
                LOG.debug("Link is impaired? {}", link.isImpaired());
                if (link.isImpaired() != hasImpairedLook) {
                    if (link.isImpaired()) {
                        getStyleClass().add(IMPAIRED_LINK_CENTER_STYLECLASS);
                        hasImpairedLook = true;
                    } else {
                        getStyleClass().remove(IMPAIRED_LINK_CENTER_STYLECLASS);

                        //remove and add basis style to reset look
                        getStyleClass().remove(LINK_CENTER_STYLECLASS);
                        getStyleClass().add(LINK_CENTER_STYLECLASS);
                        hasImpairedLook = false;
                    }

                }
            });

            Runnable updateGraphic = () -> {
                if (hasAutoSharedLook)
                    setGraphic(createSharedGraphic);
                else {
                    if (hasAddToSharedLook)
                        setGraphic(addTosharedGraphic);
                    else
                        setGraphic(null);
                }
            };

            ListChangeListener<String> autoShareLanListener =  change -> {
                boolean shared = !link.getAutoShareLanNames().isEmpty();
                LOG.debug("Link is auto shared? {}", shared);
                if (shared != hasAutoSharedLook) {
                    if (shared) {
                        //getStyleClass().add(SHARED_LINK_CENTER_STYLECLASS);
                        hasAutoSharedLook = true;
                    } else {
                        //getStyleClass().remove(SHARED_LINK_CENTER_STYLECLASS);

                        //remove and add basis style to reset look
                        getStyleClass().remove(LINK_CENTER_STYLECLASS);
                        getStyleClass().add(LINK_CENTER_STYLECLASS);
                        hasAutoSharedLook = false;
                    }
                    updateGraphic.run();

                }
            };
            link.getAutoShareLanNames().addListener(autoShareLanListener);
            autoShareLanListener.onChanged(null);

            ChangeListener<String> addToShareLanListener =  (observable, oldValue, newValue) -> {
                boolean shared = link.getSharedLan() != null;
                LOG.debug("Link connects to shared? {}", shared);
                if (shared != hasAddToSharedLook) {
                    if (shared) {
                        //getStyleClass().add(SHARED_LINK_CENTER_STYLECLASS);
                        hasAddToSharedLook = true;
                    } else {
                        //getStyleClass().remove(SHARED_LINK_CENTER_STYLECLASS);

                        //remove and add basis style to reset look
                        getStyleClass().remove(LINK_CENTER_STYLECLASS);
                        getStyleClass().add(LINK_CENTER_STYLECLASS);
                        hasAddToSharedLook = false;
                    }
                    updateGraphic.run();
                }
            };
            link.sharedLanProperty().addListener(addToShareLanListener);
            addToShareLanListener.changed(null, null, null);
            updateGraphic.run();
        }

        private void updateLinkText() {
            boolean multiSiteLink = new HashSet<>(link.getComponentManagerUrns()).size() > 1;

            if (link.getLinkTypes().isEmpty() || link.getLinkTypes().equals(Collections.singletonList("lan"))) {
                if (multiSiteLink)
                    setText(link.getClientId()+" (stitched)");
                else
                    setText(link.getClientId());
            } else {
                String typeList = link.getLinkTypes().stream()
                        .map((s) -> s.replace("-tunnel", ""))
                        .collect(Collectors.joining(", "));
                if (multiSiteLink)
                    typeList = "multi-site "+typeList;
                setText(link.getClientId()+" ("+ typeList+ ")");
            }
        }

        @Nonnull
        public FXRspecLink getRspecLink() {
            return link;
        }

        private void onLinkCenterClicked() {
            experimentCanvas.getSelectionProvider().setSelectedObject(RspecCanvasLink.this);
        }
        @Nonnull
        public RspecCanvasLink getRspecCanvasLink() {
            return RspecCanvasLink.this;
        }

        @Override
        @Nonnull
        public Point2D getCenterPosition() {
            return new Point2D(centerX, centerY);
        }

        @Override
        @Nonnull
        public Bounds getCanvasBounds() {
            return ExperimentCanvas.calculateBoundsInCanvas(linkCenter);
        }
    }


    public class InterfaceLink extends CanvasLink {
        @Nonnull
        private final FXRspecInterface rspecInterface;

        public InterfaceLink(@Nonnull FXRspecInterface rspecInterface, @Nonnull RspecCanvasNode nodeA, @Nonnull Button nodeB) {
            super(nodeA, nodeB);
            this.rspecInterface = rspecInterface;
        }

        @Override
        @Nonnull
        public Button getNodeB() {
            return (Button) super.getNodeB();
        }

        @Override
        @Nonnull
        public RspecCanvasNode getNodeA() {
            return (RspecCanvasNode) super.getNodeA();
        }

        @Nonnull
        public FXRspecInterface getRspecInterface() {
            return rspecInterface;
        }


    }
}
